def faker_anonymise(
    generator, *args, preserve_nulls=False, unique=False, **kwargs
):
    def anonymise(*, pii_value, fake, **_kwargs):
        if preserve_nulls and pii_value is None:
            return None

        faker_generator = getattr(fake.unique if unique else fake, generator)
        return faker_generator(*args, **kwargs)

    return anonymise


def preserve_internal(alternative):
    def anonymise(obj, field, pii_value, **kwargs):
        if getattr(obj, "is_superuser", False) or getattr(
            obj, "is_staff", False
        ):
            return pii_value
        return alternative(obj=obj, field=field, pii_value=pii_value, **kwargs)

    return anonymise


def const(value, preserve_nulls=False):
    def anonymise(*_, pii_value, **_kwargs):
        if preserve_nulls and pii_value is None:
            return None
        return value

    return anonymise


def random_foreign_key(obj, field, **_kwargs):
    rel = obj._meta.fields[field].related_model.objects.order_by("?").get()
    return rel.pk



from django.db.utils import DEFAULT_DB_ALIAS
from django.core.management.base import (
    BaseCommand,
    CommandError,
    CommandParser,
)

from ...engine import export_data, export_schema, validate_strategies


class Command(BaseCommand):
    help = "Export data for creating a new database."

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "only",
            nargs="*",
            help="Only export specified models.",
            metavar="app_label.ModelName",
        )
        parser.add_argument(
            "--database",
            help="The database name to sync.",
            default=DEFAULT_DB_ALIAS,
        )
        parser.add_argument(
            "--no-update",
            help="Skip updates that already exist and are non-empty.",
            action="store_true",
        )

    def handle(self, *, only=None, database, no_update, **options):
        try:
            validate_strategies(only)
        except AssertionError as e:
            raise CommandError(e)

        export_schema(database)
        export_data(database, only, no_update)



import faker
from django.conf import settings
from django.core.serializers.json import Serializer as JSONSerializer

from .utils import to_app_model_label


class PiiAnonymisingSerializer(JSONSerializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fake = faker.Faker(locale=settings.DEVDATA_FAKER_LOCALES)

    def get_dump_object(self, obj):
        data = super().get_dump_object(obj)

        for field, value in data["fields"].items():
            if field in settings.DEVDATA_FIELD_ANONYMISERS:
                data["fields"][field] = settings.DEVDATA_FIELD_ANONYMISERS[
                    field
                ](
                    obj=obj,
                    field=field,
                    pii_value=value,
                    fake=self.fake,
                )

            app_model_label = to_app_model_label(obj.__class__)
            model_anonymisers = settings.DEVDATA_MODEL_ANONYMISERS.get(
                app_model_label,
                {},
            )

            if field in model_anonymisers:
                data["fields"][field] = model_anonymisers[field](
                    obj=obj,
                    field=field,
                    pii_value=value,
                    fake=self.fake,
                )

        return data



import subprocess

from django.conf import settings
from django.core.management.color import no_style
from django.db import connections

from .strategies import Exportable
from .utils import (
    get_all_models,
    get_pg_connection_args,
    migrations_file_path,
    progress,
    psql,
    schema_file_path,
    sort_model_strategies,
    to_app_model_label,
    to_model,
)


def validate_strategies(only=None):
    not_found = []

    for model in get_all_models():
        if model._meta.abstract:
            continue

        app_model_label = to_app_model_label(model)

        if app_model_label not in settings.DEVDATA_STRATEGIES:
            if only and app_model_label not in only:
                continue

            not_found.append(app_model_label)

    if not_found:
        raise AssertionError(
            "\n".join(
                [
                    "Found models without strategies for local database creation:\n",
                    *[
                        "  * {}".format(app_model_label)
                        for app_model_label in not_found
                    ],
                ]
            )
        )


def export_schema(django_dbname):
    db_conf = settings.DATABASES[django_dbname]

    schema_file_path().parent.mkdir(exist_ok=True)
    with schema_file_path().open("w") as f:
        export_command = [
            *settings.DEVDATA_PGDUMP_COMMAND.split(),
            db_conf["NAME"],
            "--schema-only",
            "--format=plain",
            *get_pg_connection_args(db_conf),
        ]

        subprocess.run(export_command, stdout=f, check=True)

    migrations_file_path().parent.mkdir(exist_ok=True)
    with migrations_file_path().open("w") as f:
        export_command = [
            *settings.DEVDATA_PGDUMP_COMMAND.split(),
            db_conf["NAME"],
            "--data-only",
            "--table=django_migrations",
            "--format=plain",
            *get_pg_connection_args(db_conf),
        ]

        subprocess.run(export_command, stdout=f, check=True)


def export_data(django_dbname, only=None, no_update=False):
    model_strategies = sort_model_strategies(settings.DEVDATA_STRATEGIES)

    for app_model_label, strategy in progress(model_strategies):
        if only and app_model_label not in only:
            continue

        model = to_model(app_model_label)

        if isinstance(strategy, Exportable):
            strategy.export(django_dbname, model, no_update)


def sync_schema(django_dbname):
    db_conf = settings.DATABASES[django_dbname]
    pg_dbname = db_conf["NAME"]

    psql("DROP DATABASE IF EXISTS {}".format(pg_dbname), None, db_conf)
    psql(
        """
        CREATE DATABASE {database} WITH
            TEMPLATE = template0
            ENCODING = 'UTF-8'
            LC_COLLATE = 'en_GB.UTF-8'
            LC_CTYPE = 'en_GB.UTF-8'
            OWNER = {owner}
        """.format(
            owner=db_conf.get("USER", "postgres"),
            database=pg_dbname,
        ),
        None,
        db_conf,
    )

    with schema_file_path().open() as f:
        psql(f.read(), pg_dbname, db_conf)

    with migrations_file_path().open() as f:
        psql(f.read(), pg_dbname, db_conf)


def sync_data(django_dbname):
    model_strategies = sort_model_strategies(settings.DEVDATA_STRATEGIES)
    for app_model_label, strategy in model_strategies:
        model = to_model(app_model_label)

        print("Syncing {} ({})".format(app_model_label, strategy.name))
        strategy.sync(django_dbname, model)


def sync_cleanup(django_dbname):
    print("Resetting sequences")
    conn = connections[django_dbname]
    with conn.cursor() as cursor:
        for reset_sql in conn.ops.sequence_reset_sql(
            no_style(),
            get_all_models(),
        ):
            # TODO: Handle Zendesk sequence not starting at 1
            cursor.execute(reset_sql)



import pathlib
import functools
import subprocess

import tqdm

from django.apps import apps
from django.conf import settings

from django.db.models import Model


@functools.lru_cache(maxsize=1024)
def to_app_model_label(model: Model) -> str:
    return "{}.{}".format(model._meta.app_label, model.__name__)


@functools.lru_cache(maxsize=1024)
def to_model(app_model_label: str) -> Model:
    app_label, model_name = app_model_label.split(".")
    return apps.get_model(app_label, model_name)


def get_all_models():
    return apps.get_models(include_auto_created=True)


def get_pg_connection_args(db_conf):
    args = []

    for conf, arg in {
        "HOST": "host",
        "PORT": "port",
        "USER": "username",
        "PASSWORD": "password",
    }.items():
        if db_conf[conf]:
            args.append("--{}={}".format(arg, db_conf[conf]))

    return args


def psql(command, pg_dbname, connection_settings):
    psql_command = [
        *settings.DEVDATA_PSQL_COMMAND.split(),
        pg_dbname or "postgres",
        "-v",
        "ON_ERROR_STOP=1",
        *get_pg_connection_args(connection_settings),
    ]

    try:
        subprocess.run(
            psql_command,
            input=command.encode("utf-8"),
            check=True,
            stdout=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        raise AssertionError("Process failed") from None


def schema_file_path():
    return pathlib.Path(settings.DEVDATA_LOCAL_DIR) / "schema.sql"


def migrations_file_path():
    return pathlib.Path(settings.DEVDATA_LOCAL_DIR) / "migrations.sql"


def progress(sequence):
    return tqdm.tqdm(sequence)


def sort_model_strategies(model_strategies):
    model_dependencies = []
    models = set()

    for app_model_label, strategies in model_strategies.items():
        model = to_model(app_model_label)
        models.add(model)

        if hasattr(model, "natural_key"):
            deps = getattr(model.natural_key, "dependencies", [])
            if deps:
                deps = [apps.get_model(dep) for dep in deps]
        else:
            deps = []

        for field in model._meta.fields:
            if field.remote_field and field.remote_field.model != model:
                deps.append(field.remote_field.model)

        for field in model._meta.many_to_many:
            if (
                field.remote_field.through._meta.auto_created
                and field.remote_field.model != model
            ):
                deps.append(field.remote_field.model)

        for strategy in strategies:
            for dep in getattr(strategy, "depends_on", ()):
                deps.append(to_model(dep))

        model_dependencies.append((model, deps))

    model_dependencies.reverse()

    model_list = []
    while model_dependencies:
        skipped = []
        changed = False
        while model_dependencies:
            model, deps = model_dependencies.pop()

            # If all of the models in the dependency list are either already
            # on the final model list, or not on the original serialization list,
            # then we've found another model with all it's dependencies satisfied.
            if all(d not in models or d in model_list for d in deps):
                model_list.append(model)
                changed = True
            else:
                skipped.append((model, deps))
        if not changed:
            raise RuntimeError(
                "Can't resolve dependencies for {} in serialized app list.".format(
                    ", ".join(
                        "{}.{}".format(
                            model._meta.app_label, model._meta.object_name
                        )
                        for model, _ in sorted(
                            skipped, key=lambda obj: obj[0].__name__
                        )
                    ),
                ),
            )

        model_dependencies = skipped

    return [
        (to_app_model_label(y), x)
        for y in model_list
        for x in model_strategies[to_app_model_label(y)]
    ]



import argparse
import json
import sys

from django.apps import apps
from django.core.management.base import BaseCommand, CommandError
from django.utils.module_loading import import_string

from ...strategies import QuerySetStrategy


class Command(BaseCommand):
    help = "Dump the data generated by a given strategy."

    def add_arguments(self, parser):
        parser.add_argument(
            "model_label",
            metavar="app_label[.ModelName]",
            help="Restricts dumped data to the specified app_label or app_label.ModelName.",
        )
        parser.add_argument(
            "strategy_class",
            metavar="foo.bar.MyStrategy",
            help="Export strategy class.",
        )
        parser.add_argument(
            "database",
            help="The database name to sync.",
        )
        parser.add_argument(
            "kwargs",
            nargs="?",
            type=argparse.FileType("r"),
            default=sys.stdin,
            help="Path to file containing JSON encoded kwargs for the strategy",
        )

    def handle(self, model_label, strategy_class, kwargs, database, **options):
        app_name, model_name = model_label.split(".")
        app_config = apps.get_app_config(app_name)
        model = app_config.get_model(model_name)

        strategy_kwargs = json.load(kwargs)
        strategy = import_string(strategy_class)(**strategy_kwargs)

        if not isinstance(strategy, QuerySetStrategy):
            raise CommandError("Strategy not locally exportable")

        strategy.export_local(database, model, self.stdout)



import socket

from django.conf import settings
from django.db.utils import DEFAULT_DB_ALIAS
from django.core.management.base import BaseCommand, CommandError

from ...engine import sync_data, sync_schema, sync_cleanup, validate_strategies


class Command(BaseCommand):
    help = "Create new database and sync data into it."

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            help="The database name to sync.",
            default=DEFAULT_DB_ALIAS,
        )

    def handle(self, database, **options):
        try:
            validate_strategies()
        except AssertionError as e:
            raise CommandError(e)

        if (
            input(
                "You're about to delete the database {} ({}) from the host {}. "
                "Are you sure you want to continue? [y/N]: ".format(
                    self.style.WARNING(database),
                    self.style.WARNING(settings.DATABASES[database]["NAME"]),
                    self.style.WARNING(socket.gethostname()),
                ),
            ).lower()
            != "y"
        ):
            raise CommandError("Aborted")

        sync_schema(database)
        sync_data(database)
        sync_cleanup(database)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['devdata', 'devdata.management', 'devdata.management.commands']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'django-devdata',
    'version': '0.1.4',
    'description': '',
    'long_description': '# django-devdata\n\n`django-devdata` provides a convenient workflow for creating development\ndatabases seeded with anonymised production data. Have a development database\nthat contains useful data, and is fast to create and keep up to date.\n\n#### Problem\n\nIn the same way that development environments being close in configuration to\nproduction environments, it\'s important that the data in databases we use for\ndevelopment is a realistic representation of that in production.\n\nOne option is to use a dump of a production database, but there are several\nproblems with this:\n\n1. This is bad for user privacy, and therefore a security risk. It may not be\n   allowed in some organisations.\n2. It\'s a limiting factor once the production database is too big to fit on a\n   development computer.\n3. Processes to take a sample of data from a database need to preserve\n   referential integrity.\n4. It limits test data to the data available in production.\n\nAnother option is to use factories or fake data to generate the entire\ndevelopment database. This is mostly desirable, but...\n\n- It can be a burden to maintain factories once there are hundreds or thousands\n  of them.\n- It can be hard to retroactively add these to a Django site of a significant\n  size.\n\n#### Solution\n\n`django-devdata` provides defines a three step workflow:\n\n1. _Exporting_ data, with a customisable export strategies per model.\n2. _Anonymising_ data, with customisable anonymisation per field/model.\n3. _Importing_ data, with customisable importing per model.\n\n`django-devdata` ships with built-in support for:\n\n- Exporting full tables\n- Exporting subsets (random, latest, specified primary keys)\n- Anonymising data with [`faker`](https://github.com/joke2k/faker/)\n- Importing exported data\n- Importing data from [`factory-boy`](https://github.com/FactoryBoy/factory_boy)\n  factories\n\nExporting, anonymising, and importing, are all configurable, and\n`django-devdata`\'s base classes will help do this without much work.\n\n## Workflow\n\n#### Exporting\n\nThis step allows a sync strategy to persist some data that will be used to\ncreate a new development database. For example, the `QuerySetStrategy` can\nexport data from a table to a filesystem for later import.\n\nThis can be used for:\n\n- Exporting a manually created database for other developers to use.\n- Exporting realistic data from a production database.\n- A cron job to maintain a development dataset hosted on cloud storage.\n\nThis step is optional (the built-in factory strategy doesn\'t do this).\n\n#### Anonymisation\n\nThis step is critical when using `django-devdata` to export from production\nsources. It\'s not a distinct step, but rather an opt-out part of the export\nstep.\n\n#### Importing\n\nThis step is responsible for creating a new database and filling it. If any\nexporting strategies have been used those must have run first, or their outputs\nmust have been downloaded if they are being shared/hosted somewhere.\n\nFactory-based strategies generate data during this process.\n\n## Customising\n\n#### Strategies\n\nThe `django-devdata` strategies define how an import and optionally an export\nhappen. Each model is configured with a list of Strategies to use.\n\nClasses are provided to inherit from for customising this behaviour:\n\n- `Strategy` – the base class of all strategies.\n- `Exportable` –\xa0a mixin that opts this strategy in to the export step.\n- `QuerySetStrategy` – the base of all strategies that export production data\n  to a filesystem. Handles referential integrity, serialisation, and\n  anonymisation of the data pre-export.\n- `FactoryStrategy` – the base of all strategies that create data based on\n  `factory-boy` factories.\n\nThe API necessary for classes to implement is small, and there are customisation\npoints provided for common patterns.\n\nIn our experience most models can be exported with just the un-customised\n`QuerySetStrategy`, some will need to use other pre-provided strategies, and\na small number will need custom exporters based on the classes provided.\n\n#### Anonymisers\n\nAnonymisers are configured by field name, and by model and field name.\n\nEach anonymiser is a function that takes a number of kwargs with useful context\nand returns a new value, compatible with the Django JSON encoder/decoder.\n\nThe signature for an anonymiser is:\n\n```python\ndef anonymise(*, obj: Model, field: str, pii_value: Any, fake: Faker) -> Any:\n    ...\n```\n\nThere are several anonymisers provided to use or to build off:\n\n- `faker_anonymise` – Use `faker` to anonymise this field with the provided\n  generator, e.g. `faker_anonymise(\'pyint\', min_value=15, max_value=85)`.\n- `const` – anonymise to a constant value, e.g. `const(\'ch_XXXXXXXX\')`.\n- `random_foreign_key` –\xa0anonymise to a random foreign key.\n\n`django-devdata`\'s anonymisation is not intended to be perfect, but rather to be\na reasonable default for creating useful data that does a good enough job by\ndefault. _Structure_ in data can be used to de-anonymise users in some cases\nwith advanced techniques, and `django-devdata` does not attempt to solve for\nthis case as most attackers, users, and legislators, are more concerned about\nobviously personally identifiable information such as names and email addresses.\nThis anonymisation is no replacement for encryption at-rest with tools like\nFileVault or BitLocker on development machines.\n\nAn example of this pragmatism in anonymisation is the `preserve_nulls` argument\ntaken by some built-in anonymisers. This goes against _true_ anonymisation, but\nthe absence of data is typically not of much use to attackers (or concern for\nusers), if the actual data is anonymised, while this can be of huge benefit to\ndevelopers in maintaining data consistency.\n\n#### Settings\n\n`django-devdata` makes heavy use of Django settings for both defining how it\nshould act for your site, and also for configuring how you\'ll use your workflow.\n\n```python\n"""\ndjango-devdata default settings, with documentation on usage.\n"""\n\n# Required\n# A mapping of app model label to list of strategies to be used.\nDEVDATA_STRATEGIES = ...\n# {\'auth.User\': [QuerySetStrategy(name=\'all\')], \'sessions.Session\': []}\n\n# Optional\n# A mapping of field name to an anonymiser to be used for all fields with that\n# name.\nDEVDATA_FIELD_ANONYMISERS = {}\n# {\'first_name\': faker_anonymise(\'first_name\'), \'ip\': const(\'127.0.0.1\')}\n\n# Optional\n# A mapping of app model label to a mapping of fields and anonymisers to be\n# scoped to just that model.\nDEVDATA_MODEL_ANONYMISERS = {}\n# {\'auth.User\': {\'first_name\': faker_anonymise(\'first_name\')}}\n\n# Required if using any exportable strategies.\n# The full path to the directory in which to store any exported data.\nDEVDATA_LOCAL_DIR = ...\n\n# Required\n# Command to run for psql, used for importing only. If psql is on the path in\n# your development environment no change is needed.\nDEVDATA_PSQL_COMMAND = \'psql\'\n\n# Required\n# Command to run for pg_dump, used for exporting the schema and migrations only.\n# If running the export on the same machine as the database this can be left\n# unchanged, but typically this will need changing.\nDEVDATA_PGDUMP_COMMAND = \'pg_dump\'\n# \'pg_dump -h my-database-host.local -u my-user\'\n# \'ssh me@production pg_dump\'\n\n# Optional\n# Command to run for QuerySetStrategy if exporting from another machine or\n# directory. If `None`, by default, no extra process is created, but if\n# provided, this command is used in a sub process. If overridden, this will\n# typically be a call to `devdata_dump`, but that management command may be\n# overridden itself so as long as the same arguments can be passed this can be\n# anything.\nDEVDATA_DUMP_COMMAND = None\n# \'ssh me@production /opt/my-app/manage.py devdata_dump\'\n\n# Optional\n# List of locales to be used for Faker in generating anonymised data.\nDEVDATA_FAKER_LOCALES = None\n# [\'en_GB\', \'en_AU\']\n```\n',
    'author': 'Dan Palmer',
    'author_email': 'dan@danpalmer.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/danpalmer/django-devdata',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
}


setup(**setup_kwargs)



import json
import pathlib
import subprocess
from tempfile import TemporaryFile
from typing import Set, Tuple

from django.conf import settings
from django.core import serializers
from django.db import models

from .pii_anonymisation import PiiAnonymisingSerializer
from .utils import to_app_model_label, to_model


class Strategy:
    """
    Base strategy defining how to get data for a given model into a fresh
    database.
    """

    def __init__(self):
        pass

    def sync(self, django_dbname, model):
        """Load data into newly created database."""
        raise NotImplementedError


class Exportable:
    """
    A strategy that uses exported data from some source, typically a production
    database.
    """

    seen_names = set()  # type: Set[Tuple[str, str]]

    def __init__(self, *args, name, **kwargs):
        super().__init__(*args, **kwargs)

        self.name = name

    def export(self, django_dbname, model, no_update=False):
        """
        Export the data to a directory on disk. `no_update` indicates not to
        update if there is any data already existing locally.
        """
        pass

    def data_file(self, app_model_label):
        return (
            pathlib.Path(settings.DEVDATA_LOCAL_DIR)
            / app_model_label
            / "{}.json".format(self.name)
        )

    def ensure_dir_exists(self, app_model_label):
        unique_key = (app_model_label, self.name)
        if unique_key in self.seen_names:
            raise ValueError(
                "Exportable strategy names must be unique per model so that "
                "exports do not collide.",
            )
        self.seen_names.add(unique_key)

        model_dir = pathlib.Path(settings.DEVDATA_LOCAL_DIR) / app_model_label
        model_dir.mkdir(exist_ok=True)


class QuerySetStrategy(Exportable, Strategy):
    """Sync a table using an unfiltered QuerySet."""

    use_natural_foreign_keys = False
    use_natural_primary_keys = False

    json_indent = 2

    def __init__(self, *args, restricted_pks=None, anonymise=True, **kwargs):
        super().__init__(*args, **kwargs)

        # Only set in the remote dump process
        self.restricted_pks = restricted_pks
        self.anonymise = anonymise

    def get_kwargs(self, model):
        restricted_pks = {}

        for field in model._meta.fields:
            if not field.related_model:
                continue

            app_model_label = to_app_model_label(field.related_model)
            restricted_pks[app_model_label] = self.get_exported_pks_for_model(
                field.related_model,
            )

        return {
            "name": self.name,
            "anonymise": self.anonymise,
            "restricted_pks": restricted_pks,
        }

    def get_queryset(self, django_dbname, model):
        queryset = model.objects.using(django_dbname)

        for app_model_label, restrict_pks in self.restricted_pks.items():
            restrict_model = to_model(app_model_label)

            # We filter to all fields that relate to the restricted model. This
            # will usually be just one field, but in cases where it's multiple
            # to preserve FK integrity we must restrict all of them.
            fk_fields = [
                x
                for x in model._meta.fields
                if x.related_model == restrict_model
                if x.related_model != model
            ]

            queryset = queryset.filter(
                *[
                    models.Q(
                        **{x.attname: None},
                    )
                    | models.Q(
                        **{"{}__in".format(x.attname): restrict_pks},
                    )
                    for x in fk_fields
                ]
            )

        return queryset

    def export_local(self, django_dbname, model, output):
        queryset = self.get_queryset(django_dbname, model)

        serializer = (
            PiiAnonymisingSerializer()
            if self.anonymise
            else serializers.get_serializer("json")
        )

        serializer.serialize(
            queryset.iterator(),
            indent=self.json_indent,
            use_natural_foreign_keys=self.use_natural_foreign_keys,
            use_natural_primary_keys=self.use_natural_primary_keys,
            stream=output,
        )

    def export(self, django_dbname, model, no_update):
        app_model_label = to_app_model_label(model)
        data_file = self.data_file(app_model_label)

        if no_update and data_file.exists():
            return

        kwargs = self.get_kwargs(model)

        command = settings.DEVDATA_DUMP_COMMAND.split()
        command.append(app_model_label)
        command.append(
            "{}.{}".format(self.__class__.__module__, self.__class__.__name__)
        )
        command.append(django_dbname)

        self.ensure_dir_exists(app_model_label)

        # We pass the kwargs over stdin because if we have large querysets in
        # the upstream dependencies of this there could be too much data to pass
        # in the arguments to exec.
        with TemporaryFile() as temp_f, TemporaryFile("wt") as kwargs_in_f:
            json.dump(kwargs, kwargs_in_f)
            kwargs_in_f.seek(0)

            try:
                subprocess.run(
                    command,
                    stdin=kwargs_in_f,
                    stdout=temp_f,
                    check=True,
                )
            except subprocess.CalledProcessError:
                print(
                    "Failed to export {} ({})".format(
                        app_model_label,
                        self.name,
                    )
                )
                raise

            temp_f.seek(0)

            with data_file.open("wb") as f:
                written = f.write(temp_f.read())

        # Check if we got an empty JSON list, but assume that the file size will
        # be small if so, to prevent reading in huge data files.
        if written < 1000:
            with data_file.open("r") as f:
                if json.load(f) == []:
                    print(
                        "Warning! '{}' exporter for {} selected no data.".format(
                            self.name,
                            app_model_label,
                        )
                    )

    def sync(self, django_dbname, model):
        with self.data_file(to_app_model_label(model)).open() as f:
            objects = serializers.deserialize("json", f, using=django_dbname)
            self.sync_objects(django_dbname, model, objects)

    def sync_objects(self, django_dbname, model, objects):
        qs = model.objects.using(django_dbname)
        existing_pks = set(qs.values_list("pk", flat=True))
        qs.bulk_create(
            [x.object for x in objects if x.object.pk not in existing_pks]
        )

    def get_exported_pks_for_model(self, model):
        return [
            str(x["pk"]) for x in self.get_exported_objects_for_model(model)
        ]

    def get_exported_objects_for_model(self, model):
        app_model_label = to_app_model_label(model)
        objects = []

        data_dir = pathlib.Path(settings.DEVDATA_LOCAL_DIR) / app_model_label
        data_files = data_dir.glob("*.json")

        for data_file in data_files:
            with data_file.open() as f:
                try:
                    data = json.load(f)
                except json.JSONDecodeError:
                    print("Invalid file {}".format(data_file))
                    raise

                objects.extend(data)

        return objects


class ExactQuerySetStrategy(QuerySetStrategy):
    """Sync specific rows from a table using a QuerySet filtered to given PKs."""

    def __init__(self, *args, pks, **kwargs):
        super().__init__(*args, **kwargs)
        self.pks = pks

    def get_kwargs(self, model):
        return {**super().get_kwargs(model), "pks": self.pks}

    def get_queryset(self, django_dbname, model):
        return (
            super().get_queryset(django_dbname, model).filter(pk__in=self.pks)
        )


class RandomSampleQuerySetStrategy(QuerySetStrategy):
    """Syncs a random sample from a QuerySet."""

    def __init__(self, *args, count, **kwargs):
        super().__init__(*args, **kwargs)

        self.count = count

    def get_kwargs(self, model):
        return {**super().get_kwargs(model), "count": self.count}

    def get_queryset(self, django_dbname, model):
        return (
            super()
            .get_queryset(django_dbname, model)
            .order_by("?")[: self.count]
        )


class LatestSampleQuerySetStrategy(QuerySetStrategy):
    """Syncs the latest items from a QuerySet."""

    def __init__(self, *args, count, order_by="-id", **kwargs):
        super().__init__(*args, **kwargs)

        self.count = count
        self.order_by = order_by

    def get_kwargs(self, model):
        return {
            **super().get_kwargs(model),
            "count": self.count,
            "order_by": self.order_by,
        }

    def get_queryset(self, django_dbname, model):
        qs = super().get_queryset(django_dbname, model)
        return qs.order_by(self.order_by)[: self.count]


class ModelReverseRelationshipQuerySetStrategy(QuerySetStrategy):
    """
    Base class for inverse relationship queryset exporting.
    Useful for relationships that go the wrong way to work with
    normal dependency analysis. Example:

        class Order(Model):
            user = ForeignKey(User)
            charge = ForeignKey(Charge)

        class Charge(Model):
            ...

    Requires a dependent model, such as `User`, and defines an additional
    explicit dependency on that model locally. Reads in those IDs, passes them
    to the exporting process, where they are available to `get_queryset`.
    """

    depends_on = NotImplemented  # type: Tuple[str, ...]

    def __init__(self, *args, reverse_filter=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.reverse_filter = reverse_filter

    def get_kwargs(self, model):
        return {
            **super().get_kwargs(model),
            "reverse_filter": self.get_reverse_filter(model),
        }

    def get_reverse_filter(self, model):
        raise NotImplementedError

    def get_queryset(self, django_dbname, model):
        qs = super().get_queryset(django_dbname, model)
        return qs.filter(**self.reverse_filter)


class FactoryStrategy(Strategy):
    """
    Use the provided factory/factories to create data for this model (and any
    related) models.
    """

    def __init__(self, *args, factories, **kwargs):
        super().__init__(*args, **kwargs)
        self.factories = factories

    def sync(self, django_dbname, model):
        pass
